# Copyright 2022 PCL, Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""
transform wikitext-2, wikitext-103, lambada, openwebtext dataset to mindrecord.
"""
import argparse
import glob
import json
import os
import re
from functools import partial
from multiprocessing import Pool, current_process
import numpy as np

from mindspore.mindrecord import FileWriter
from src.tokenization_jieba import JIEBATokenizer




def chunks(lst, n):
    """ yield n sized chunks from list"""
    for i in range(0, len(lst), n):
        yield lst[i:i+n]


def package_file(it, n):
    """ package multiple files"""
    stop = False
    while not stop:
        batch = []
        for _ in range(n):
            try:
                batch.append(next(it))
            except StopIteration:
                stop = True
        if not batch:
            break
        yield batch


def clean_wikitext(string):
    """ cleaning wikitext dataset"""
    # contractions
    string = string.replace("s '", "s'")
    string = re.sub(r"/' [0-9]/", r"/'[0-9]/", string)
    # number separators
    string = string.replace(" @-@ ", "-")
    string = string.replace(" @,@ ", ",")
    string = string.replace(" @.@ ", ".")
    # punctuation
    string = string.replace(" : ", ": ")
    string = string.replace(" ; ", "; ")
    string = string.replace(" . ", ". ")
    string = string.replace(" ! ", "! ")
    string = string.replace(" ? ", "? ")
    string = string.replace(" , ", ", ")
    # double brackets
    string = re.sub(r"\(\s*([^\)]*?)\s*\)", r"(\1)", string)
    string = re.sub(r"\[\s*([^\]]*?)\s*\]", r"[\1]", string)
    string = re.sub(r"{\s*([^}]*?)\s*}", r"{\1}", string)
    string = re.sub(r"\"\s*([^\"]*?)\s*\"", r'"\1"', string)
    string = re.sub(r"'\s*([^']*?)\s*'", r"'\1'", string)
    # miscellaneous
    string = string.replace("= = = =", "====")
    string = string.replace("= = =", "===")
    string = string.replace("= =", "==")
    string = string.replace(" "+chr(176)+" ", chr(176))
    string = string.replace(" \n", "\n")
    string = string.replace("\n ", "\n")
    string = string.replace(" N ", " 1 ")
    string = string.replace(" 's", "'s")
    return string


def tokenize_openwebtext(tokenizer, iterator, seq_length, eot):
    """ tokenize openwebtext dataset"""
    for file_path in iterator:
        if os.path.getsize(file_path) == 0:
            continue
        content = []
        with open(file_path, 'r', encoding='utf-8') as f:
            for para in f.read().split("\n\n"):
                if para:
                    tokenized_text = tokenizer.tokenize(para)
                    content += tokenizer.convert_tokens_to_ids(tokenized_text) + [
                        eot]
        for chunk in chunks(content, seq_length):
            sample = {}
            if len(chunk) == seq_length:
                sample['input_ids'] = np.array(chunk, dtype=np.int32)
                yield sample


def tokenize_wiki(tokenizer, file_path, seq_length, eot):
    """tokenize wikitext-2/wikitext-103 dataset"""
    content = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for para in clean_wikitext(f.read()).split("\n\n"):
            if para and para.strip().startswith('=') is False:
                tokenized_text = tokenizer.tokenize(para)
                content += tokenizer.convert_tokens_to_ids(tokenized_text) + [
                    eot]
    for chunk in chunks(content, seq_length):
        sample = {}
        if len(chunk) == seq_length:
            sample['input_ids'] = np.array(chunk, dtype=np.int32)
            yield sample


def tokenize_lambada(tokenizer, file_path, seq_length, eot):
    """tokenize lambada dataset"""
    content = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for line in f.readlines():
            para = json.loads(line)['text'].replace(
                "“", '"').replace("”", '"').strip().strip(".")
            tokenized_text = tokenizer.tokenize(para)
            content += tokenizer.convert_tokens_to_ids(tokenized_text) + [eot]
    for chunk in chunks(content, seq_length):
        sample = {}
        if len(chunk) == seq_length:
            sample['input_ids'] = np.array(chunk, dtype=np.int32)
            yield sample


def task_unit(iterator, tokenizer, seq_length, eot, parallel_writer=True):
    """task for each process"""
    p = current_process()
    index = p.pid if p.pid else 0

    item_iter = tokenize_openwebtext(tokenizer, iterator, seq_length, eot)
    batch_size = 1024  # size of write batch
    count = 0
    while True:
        data_batch = []
        try:
            for _ in range(batch_size):
                data_batch.append(next(item_iter))
                count += 1
            writer.write_raw_data(data_batch, parallel_writer=parallel_writer)
            print("Process {} transformed {} records.".format(
                index, count))
        except StopIteration:
            if data_batch:
                writer.write_raw_data(data_batch,
                                      parallel_writer=parallel_writer)
                print("Process {} transformed {} records.".format(
                    index, count))
            break


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset_type', type=str, default='openwebtext')
    parser.add_argument('--input_glob', type=str, default='*.txt')
    parser.add_argument('--output_file', type=str,
                        default='./output/transfered_mindrecord')
    parser.add_argument('--tokenizer', type=str, default='gpt', choices=['gpt', 'jieba'])
    parser.add_argument('--vocab_file', type=str, default=None)
    parser.add_argument('--model_file', type=str, default=None)
    parser.add_argument('--file_partition', type=int, default=1)
    parser.add_argument('--file_batch_size', type=int, default=1024)
    parser.add_argument('--num_process', type=int, default=64)
    parser.add_argument('--seq_length', type=int, default=1025)
    parser.add_argument('--eot', type=int, default=50256)
    parser.add_argument('--data_column_name', type=str, default='input_ids')


    args = parser.parse_args()

    out_dir, out_file = os.path.split(os.path.abspath(args.output_file))
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)
    schema = {args.data_column_name: {"type": "int32", "shape": [-1]},}
    writer = FileWriter(file_name=args.output_file,
                        shard_num=args.file_partition)
    writer.add_schema(schema, args.dataset_type)
    writer.open_and_set_header()

    # Start to load tokenizer
    if args.tokenizer == 'gpt':
        try:
            from transformers import GPT2Tokenizer
        except ModuleNotFoundError:
            print("module 'transformers' not installed.")
        word_tokenizer = GPT2Tokenizer.from_pretrained("gpt2")
    else:
        if not os.path.exists(args.vocab_file):
            raise FileNotFoundError(f"file {args.vocab_file} do not exists.")
        if not os.path.exists(args.model_file):
            raise FileNotFoundError(f"file {args.model_file} do not exists.")
        word_tokenizer = JIEBATokenizer(vocab_file=args.vocab_file, model_file=args.model_file)

    transforms_count = 0
    if args.dataset_type == 'wiki':
        for x in tokenize_wiki(word_tokenizer, args.input_glob, args.seq_length, args.eot):
            transforms_count += 1
            writer.write_raw_data([x])
        print("Transformed {} records.".format(transforms_count))
    elif args.dataset_type == 'lambada':
        for x in tokenize_lambada(word_tokenizer, args.input_glob, args.seq_length, args.eot):
            transforms_count += 1
            writer.write_raw_data([x])
        print("Transformed {} records.".format(transforms_count))
    elif args.dataset_type == 'openwebtext':
        file_iter = glob.iglob(args.input_glob)
        with Pool(processes=args.num_process) as pool:
            map_func = partial(task_unit, tokenizer=word_tokenizer, seq_length=args.seq_length, eot=args.eot)
            pool.map(map_func, package_file(file_iter, args.file_batch_size))
    else:
        raise ValueError(
            "Not support dataset type: {}".format(args.dataset_type))

    writer.commit()
    out_file = args.output_file
    if args.file_partition > 1:
        out_file += '0'
    print("Transform finished, output files refer: {}".format(out_file))



'''
    code by TaeHwan Jung(@graykode)
    Original Paper and repository here : https://github.com/openai/gpt-2
    GPT2 Pytorch Model : https://github.com/huggingface/pytorch-pretrained-BERT
'''
import numpy as np
import os
from .pangu.sample import generate

SESSION = None
INT8_MODEL_PATH = None

def create_model_for_provider(model_path: str, provider: str= 'CPUExecutionProvider',threads=2):
    from onnxruntime import GraphOptimizationLevel, InferenceSession, SessionOptions, get_all_providers

    assert provider in get_all_providers(), f"provider {provider} not found, {get_all_providers()}"
    # Few properties that might have an impact on performances (provided by MS)
    options = SessionOptions()
    options.intra_op_num_threads = int(os.environ.get('NUM_THREADS', 2))
    options.graph_optimization_level = GraphOptimizationLevel.ORT_ENABLE_ALL
    # Load the model as a graph and prepare the CPU backend
    session = InferenceSession(model_path, options, providers=[provider])
    session.disable_fallback()

    return session

def onnx_infer_func(onnx_model_path: str,threads=2):
    input_names = ["input_0","past"]
    global SESSION
    if INT8_MODEL_PATH != onnx_model_path:
        SESSION = create_model_for_provider(onnx_model_path,'CPUExecutionProvider',threads=threads)
    def forward(input_ids,
                position_ids,
                token_type_ids,
                lm_labels,
                past):
        return SESSION.run(None, {
            input_names[0]: input_ids.astype(np.int32),
            input_names[1]: past,
        })
    return forward

def onnx_generate(input,onnx_int8_model_path,tokenizer,past_path,
                   topk=1,top_p=0.9,threads=2,max_len=500):
    assert os.path.exists(onnx_int8_model_path), f"{onnx_int8_model_path} not found"
    assert onnx_int8_model_path.endswith('.onnx'), f"{onnx_int8_model_path} is not a onnx model"
    assert past_path.endswith('.npy'), f"{past_path} is not a npy file"

    past=np.load(past_path)
    infer_func = onnx_infer_func(onnx_int8_model_path,threads=threads)
    txt = generate(infer_func, input, tokenizer, past,max_len=max_len,
                   top_p=top_p, top_k=topk,temperature=1.0,)

    return txt

if __name__ == '__main__':
    from pcl_pangu.tokenizer.tokenization_jieba import get_tokenizer

    tokenizer = get_tokenizer()



    # generate_ = lambda text,max_len=100: generate(infer_func, text, tokenizer, past,max_len=max_len)
    # print(generate_('青椒肉丝的做法：\n'))
    # print(generate_('西红柿炒鸡蛋的做法：\n'))
    # print(generate_('上联：天地在我心中\n下联：'))
    # print(generate_('1+1=2;3+5=8;2+4=', max_len=2))
    # print(generate_('默写古诗：\n白日依山尽，黄河入海流。\n床前明月光，', max_len=5))
    # print(generate_('李大嘴：“各回各家，各找各妈！” \n佟掌柜：'))
    # print(generate_('中国的首都是北京\n日本的首都是东京\n美国的首都是'))
    # print(generate_('中国的四大发明有哪些？', 50))
    # print(generate_('''乔布斯曾经说过：“''', 50))
    # print(generate_('''老子曾经说过：“''', 50))
    # print(generate_('''老子曾经说过：“''', 50))





from pcl_pangu.model import alpha
from pcl_pangu.context import set_context
from pcl_pangu.model.launcher_torch import launch
from loguru import logger
import os, sys


DISTRUBUTED_CONFIG = {
    'nnodes':1,
    'node_rank':0,
    'nproc_per_node':1,
    'master_addr':"localhost",
    'master_port':29503
}

def run_pt(script_args, py_script, **kwargs):
    current_dir = os.path.dirname(os.path.dirname(__file__))
    sys.path.append(current_dir+'/panguAlpha_pytorch')
    py_script = os.path.join(current_dir, py_script)
    logger.info("> Running {} with args: {}".format(py_script, script_args))
    launch(training_script=py_script,
           training_script_args=script_args,
           **DISTRUBUTED_CONFIG,
           **kwargs)

def split_pt(config, merged_path, num_ranks):
    config.save = merged_path
    config.load = merged_path
    script_args = config._get_training_script_args()

    model_save = os.path.join(merged_path, f'{num_ranks}_partition_model')
    if os.path.exists(model_save):
        os.system('rm -rf {}'.format(model_save))
    os.system('mkdir {}'.format(model_save))
    script_args.append('--num-mp-model={}'.format(num_ranks))
    script_args.append('--model-type={}'.format('Pangu'))
    script_args.append('--mp-model-save={}'.format(model_save))
    py_script = 'model/panguAlpha_pytorch/tools/split_full_model_into_mp_model.py'
    run_pt(script_args, py_script)

if __name__ == '__main__':
    from pcl_pangu.model_converter import split_pt
    merged_path = '/home/yands/tmp/numpy_ckpt_2.6b/merged'
    split_pt(alpha.model_config_gpu(model='2B6'),merged_path,2)


    pass


import os
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),
                                             os.path.pardir)))

import torch

from megatron import mpu
from megatron.checkpointing import ensure_directory_exists
from megatron.checkpointing import get_checkpoint_name
from megatron.checkpointing import get_checkpoint_tracker_filename
from megatron.global_vars import rebuild_tokenizer
from megatron.global_vars import _parse_args


def get_model(model_type):

    if model_type == 'BERT':
        from pretrain_bert import model_provider
    elif model_type == 'GPT2':
        from pretrain_gpt2 import model_provider
    elif model_type == 'Pangu':
        from pretrain_gpt2 import model_provider
    elif model_type == 'RACE':
        from tasks.race.finetune import model_provider
    elif model_type == ['MNLI', 'QQP']:
        num_classes = 2
        if model_type == 'MNLI':
            num_classes = 3
        from megatron.model.classification import Classification
        def model_provider():
            return Classification(num_classes=num_classes, num_tokentypes=2)
    else:
        raise Exception('unrecognized model type: {}'.format(model_type))

    model = model_provider()
    model = model.half()

    return model


def get_mp_split_args(parser):
    """Provide extra arguments required for merging."""
    group = parser.add_argument_group(title='mp merge')

    group.add_argument('--model-type', type=str, required=True,
                       choices=['BERT', 'GPT2', 'Pangu', 'RACE', 'MNLI', 'QQP'],
                       help='Type of the mdoel.')
    group.add_argument('--num-mp-model', type=int, required=True,
                       help='Number of mp model you want to divid into.')
    group.add_argument('--mp-model-save', type=str, required=True,
                       help='Dir of saving model parallel model.')

    return parser


def get_full_model_loadCkpt(args):
    args.model_parallel_size = 1
    tokenizer = rebuild_tokenizer(args)

    mpu.initialize.set_model_parallel_world_size(1)
    mpu.initialize.set_model_parallel_rank(0)
    model = get_model(args.model_type)


    tracker_filename = get_checkpoint_tracker_filename(args.load)
    with open(tracker_filename, 'r') as f:
        metastring = f.read().strip()
        iteration = int(metastring)
    checkpoint_name = get_checkpoint_name(args.load, iteration)
    sd = torch.load(checkpoint_name, map_location='cpu')
    model.load_state_dict(sd['model'])

    return model


def get_one_partition_model(args, rank):
    args.model_parallel_size = args.num_mp_model
    tokenizer = rebuild_tokenizer(args)

    mpu.initialize.set_model_parallel_world_size(args.num_mp_model)
    mpu.initialize.set_model_parallel_rank(rank)
    model = get_model(args.model_type)

    return model


def split_into_partitions(tensor, num_partitions, partition_dim, stride, rank):

    per_partition_size = mpu.utils.divide(tensor.size(partition_dim),
                                          num_partitions)
    per_partition_per_stride_size = mpu.utils.divide(per_partition_size, stride)

    partitions_list = torch.split(tensor,
                                  per_partition_per_stride_size,
                                  dim=partition_dim)

    partitions = []
    for i in range(num_partitions):
        partition = torch.cat(partitions_list[i::num_partitions],
                              dim=partition_dim)
        partitions.append(partition)

    return partitions[rank]


def split_full_model(full_parameter, partition_parameter, partition_dim, stride,
                     rank, num_partitions):

    per_partition_size = partition_parameter.size(partition_dim)

    def split():
        with torch.no_grad():
            if (per_partition_size * num_partitions) == full_parameter.size(
                    partition_dim):
                partition_parameter.data.copy_(
                    split_into_partitions(
                        full_parameter, num_partitions, partition_dim, stride, rank
                    ).data
                )
            else:
                assert full_parameter.size(partition_dim) \
                       <= per_partition_size * num_partitions

                dim_diff = (per_partition_size * num_partitions) - \
                           full_parameter.size(partition_dim)
                print('     ***WARNING*** sizes do not match. Will add '
                      'the merged partitions by {} along dimension {} '
                      'to increase the size from {} to {} ...'.format(
                    dim_diff,
                    partition_dim,
                    full_parameter.size(partition_dim),
                    per_partition_size * num_partitions)
                )

                size = list(full_parameter.size())
                size[partition_dim] = dim_diff
                add_tensor = torch.ones(size)
                full_parameter_ = torch.cat((full_parameter.data, add_tensor),
                                            partition_dim)

                assert full_parameter_.size(partition_dim) \
                       == per_partition_size * num_partitions

                partition_parameter.data.copy_(
                    split_into_partitions(
                        full_parameter_, num_partitions, partition_dim, stride, rank
                    ).data
                )

    # If stride is 1, then do simple concatination.
    if stride == 1:
        split()
        return
    else :
        raise ValueError("stride must be 1")
    return


def load_one_partition_model(partition_model, full_model, rank):

    full_params_gen = full_model.named_parameters()
    partition_params_gen = partition_model.named_parameters()
    while True:
        try:

            # Get the params and check names.
            name, full_param = next(full_params_gen)
            partition_name, partition_param = next(partition_params_gen)

            print(' > working on {} ...'.format(name))
            print('     merged         type: {}, size: {}'.format(
                full_param.dtype, list(full_param.size())))
            assert partition_name == name
            print('     partition {}    type: {}, size: {}'.format(
                rank, partition_param.dtype, list(partition_param.size())))

            # For the non-parallel parameters, simply copy from full model.
            if not hasattr(full_param, 'model_parallel'):
                print('     none-parallel parameter, simple copy from full model')
                with torch.no_grad():
                    partition_param.data.copy_(full_param.data)
            # For parallel parameters, split full model
            else:
                print('     split full model with stride {} along '
                      'dimention {}'.format(full_param.stride,
                                            full_param.partition_dim))

                split_full_model(full_param, partition_param,
                                 full_param.partition_dim,
                                 full_param.stride,
                                 rank,
                                 args.num_mp_model)

                # split_to_partitions(full_param,
                #                  partitions_param,
                #                  full_param.partition_dim,
                #                  full_param.stride)

        except StopIteration:
            break


if __name__ == '__main__':

    args = _parse_args(extra_args_provider=get_mp_split_args)
    save = args.mp_model_save

    iteration = 'iter_0007800'  # inter_***    any id is ok
    modelName = 'model_optim_rng.pt'
    full_model = get_full_model_loadCkpt(args)
    for rank in range(args.num_mp_model):
        partition_model = get_one_partition_model(args, rank)
        load_one_partition_model(partition_model, full_model,
                                 rank)

        sd = {}
        sd['model'] = partition_model.state_dict_for_save_checkpoint()
        sd['iteration'] = iteration
        rankName = 'mp_rank_{:02d}'.format(rank)
        savePath='/'.join([save,iteration,rankName,modelName])
        ensure_directory_exists(savePath)
        torch.save(sd, savePath)

    recordFile = open(get_checkpoint_tracker_filename(save), 'w')
    recordFile.writelines(iteration.split('_')[1])
    recordFile.close()
    print('Done!')
    print('Checkpoints saved to {}'.format(save))
    pass




# coding=utf-8
# Copyright (c) 2020, NVIDIA CORPORATION. 2022, PCL. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pretrain AlphaEnhanced"""

import torch

from megatron import get_args
from megatron import print_rank_0
from megatron import get_timers
from megatron import get_tokenizer
from megatron import mpu
from megatron.data.gpt2_dataset import build_train_valid_test_datasets
from megatron.model import EnhancedModel
from megatron.training import pretrain
from megatron.utils import get_ltor_masks_and_position_ids
from megatron.utils import reduce_losses

def model_provider():
    """Build the model."""

    print_rank_0('building AlphaEnhanced model ...')
    model = EnhancedModel(num_tokentypes=0, parallel_output=True)

    return model


def get_batch(data_iterator):
    """Generate a batch"""
    args = get_args()
    tokenizer = get_tokenizer()

    # Items and their type.
    keys = ['text']
    datatype = torch.int64

    # Broadcast data.
    if data_iterator is not None:
        data = next(data_iterator)
    else:
        data = None
    data_b = mpu.broadcast_data(keys, data, datatype)

    # Unpack.
    tokens_ = data_b['text'].long()
    labels = tokens_[:, 1:].contiguous()
    tokens = tokens_[:, :-1].contiguous()

    # Get the masks and postition ids.
    attention_mask, loss_mask, position_ids = get_ltor_masks_and_position_ids(
        tokens,
        tokenizer.eod,
        args.reset_position_ids,
        args.reset_attention_mask,
        args.eod_mask_loss)

    return tokens, labels, loss_mask, attention_mask, position_ids


def forward_step(data_iterator, model):
    """Forward step."""
    args = get_args()
    timers = get_timers()

    # Get the batch.
    timers('batch generator').start()
    tokens, labels, loss_mask, attention_mask, position_ids = get_batch(data_iterator)
    timers('batch generator').stop()
    # Forward model.
    losses, _ = model(tokens, position_ids, attention_mask, labels=labels)
    loss_mask = loss_mask.view(-1)
    loss = torch.sum(losses.view(-1) * loss_mask) / loss_mask.sum()

    # Reduce loss for logging.
    reduced_loss = reduce_losses([loss])

    return loss, {'lm loss': reduced_loss[0]}


def train_valid_test_datasets_provider(train_val_test_num_samples):
    """Build train, valid, and test datasets."""
    args = get_args()

    print_rank_0('> building train, validation, and test datasets for AlphaEnhanced ...')
    train_ds, valid_ds, test_ds = build_train_valid_test_datasets(
        data_prefix=args.data_path,
        data_impl=args.data_impl,
        splits_string=args.split,
        train_valid_test_num_samples=train_val_test_num_samples,
        seq_length=args.seq_length,
        seed=args.seed,
        skip_warmup=(not args.mmap_warmup))
    print_rank_0("> finished creating AlphaEnhanced datasets ...")

    return train_ds, valid_ds, test_ds


if __name__ == "__main__":

    pretrain(train_valid_test_datasets_provider, model_provider, forward_step,
             args_defaults={'tokenizer_type': 'GPT2BPETokenizer'})



import torch

class LayerNorm_(torch.nn.Module):
    def __init__(self, normalized_shape, init_method, dp=4, eps=1e-5, scale=1e-3):
        super(LayerNorm_, self).__init__()
        # Initialize master weight
        self.gamma = torch.empty(normalized_shape,
                                 dtype=torch.float,
                                 requires_grad=True)
        init_method(self.gamma)
        self.beta = torch.empty(normalized_shape,
                                dtype=torch.float,
                                requires_grad=True)
        init_method(self.beta)
        self.eps = eps

    def construct(self, x):
        mean = torch.mean(x, -1, keepdim=True)
        diff = torch.sub(x, mean)
        variance = torch.mean(torch.square(diff), -1,keepdim=True)
        variance_eps = torch.sqrt(torch.add(variance, self.eps))
        output_ = torch.div(diff, variance_eps)
        output = torch.add(torch.mul(output_, self.gamma), self.beta)
        return output
